@@ -17,6 +17,7 @@ import os
 import shutil
 import sys
 import subprocess
+import time
 
 import psycopg2
 import pydash
@@ -33,10 +34,21 @@ class Patch:
         self.__init_mode = init_mode
         # self.__orig_dir = os.path.abspath('.')
         self.__module_dir = os.path.dirname(__file__)
-        self.__last_release_s = None
-        self.__release = None
-        self.__release_s = ''
-        self.__release_path = None
+        self.__curr_release = None
+        self.__curr_release_s = None
+        self.__prev_release = None
+        self.__next_release = None
+        self.__update_release()
+
+    def __update_release(self):
+        self.__curr_release = self.__hop_cls.get_current_db_release()
+        self.__curr_release_s = self.__hop_cls.get_release_s(self.__curr_release)
+        self.__prev_release = self.__hop_cls.get_previous_release()
+        self.__prev_release_s = self.__hop_cls.get_release_s(self.__prev_release)
+        self.__next_release = self.__hop_cls.get_next_release()
+        self.__next_release_s = self.__next_release and self.__hop_cls.get_release_s(self.__next_release)
+        # print('XXX update release', self.__prev_release_s, self.__curr_release_s, self.__next_release_s)
+
 
     @property
     def model(self):
@@ -55,10 +67,10 @@ class Patch:
 
     @property
     def __patch_path(self):
-        return f'{self.__hop_cls.project_path}/Patches/{self.__release_path}/'
+        return f'{self.__hop_cls.project_path}/Patches/{self.__hop_cls.release_path}/'
 
     def __get_backup_file_name(self, release):
-        release_s = self.get_release_s(release)
+        release_s = self.__hop_cls.get_release_s(release)
         return f'{self.__hop_cls.project_path}/Backups/{self.dbname}-{release_s}.sql'
 
     @property
@@ -66,24 +78,22 @@ class Patch:
         with open(os.path.join(self.__patch_path, 'CHANGELOG.md'), encoding='utf-8') as changelog:
             return changelog.read()
 
-    def __get_previous_release(self):
-        "Returns the penultimate release"
-        #pylint: disable=invalid-name
-        Previous = self.__hop_cls.model.get_relation_class(
-            'half_orm_meta.view.hop_penultimate_release')
-        return next(Previous().select())
+    def is_first_release(self):
+        return self.__curr_release_s == '0.0.0'
 
-    def _revert(self):
+    def revert(self):
         """Revert to the previous release
 
         Needs the backup
         """
-        prev_release = self.__get_previous_release()
-        curr_release = self.get_current_release()
-        backup_file = self.__get_backup_file_name(prev_release)
+        self.__update_release()
+        if self.is_first_release():
+            sys.stderr.write('Current release is 0.0.0. There is no previous release to revert to!\n')
+            sys.exit(1)
+        backup_file = self.__get_backup_file_name(self.__prev_release)
         if os.path.exists(backup_file):
             self.__hop_cls.model.disconnect()
-            print("Restoring previous release...")
+            print("Restoring previous DB release...")
             try:
                 subprocess.run(['dropdb', self.dbname], check=True)
             except subprocess.CalledProcessError:
@@ -99,17 +109,33 @@ class Patch:
             #pylint: disable=invalid-name
             Release = self.__hop_cls.model.get_relation_class('half_orm_meta.hop_release')
             Release(
-                major=curr_release['major'],
-                minor=curr_release['minor'],
-                patch=curr_release['patch']
+                major=self.__curr_release['major'],
+                minor=self.__curr_release['minor'],
+                patch=self.__curr_release['patch']
                 ).delete()
-            print(f'Reverted to {self.get_release_s(prev_release)}')
+            self.__update_release()
+            print(f'Reverted to {self.__curr_release_s}')
+            self.__hop_cls.what_next()
+            # self.__hgit.set_branch(self.__curr_release_s)
         else:
-            print(f'Revert failed! No backup file for {prev_release}.')
+            print(f'Revert failed! No backup file for {prev_release_s}.')
 
     def patch(self, force=False, revert=False):
-        #TODO: docstring
-        "patch method"
+        """Patches the repo
+
+        Gets the current patch release
+        Gets the next release to apply.
+        If the branch is hop_main applies the next patch available (Patches/X/Y/Z)
+        If the method is invoked and the last applied patch correspond to the git branch, reverts and applie the patch.
+
+        Args:
+            force (bool, optional): [description]. Defaults to False.
+            revert (bool, optional): [description]. Defaults to False.
+
+        Returns:
+            [type]: [description]
+        """
+        self.__update_release()
         if self.__hop_cls.production:
             # we ensure that we are on the hop_main branch in prod
             # we set force and revert to False
@@ -120,29 +146,31 @@ class Patch:
             self.__hgit.repo.git.pull()
 
         if self.__create_mode or self.__init_mode:
-            self.__last_release_s = 'pre-patch'
+            self.__hop_cls.last_release_s = 'pre-patch'
             self.save_database()
             return self._init()
+
         if revert:
-            return self._revert()
-        branch_name = str(self.__hgit.repo.active_branch)
-        curr_release = self.get_release_s(self.get_current_release())
-        if branch_name == f'hop_{curr_release}':
-            revert_i = input(f'Replay patch {curr_release} [Y/n]? ') or 'Y'
-            if revert_i.upper() == 'Y':
-                self._revert()
-                force = True
-            else:
-                sys.exit()
-        self._patch(force=force)
-        return self.__release_s
+            self.revert()
+        else:
+            branch_name = str(self.__hgit.repo.active_branch)
+            if branch_name == f'hop_{self.__curr_release_s}':
+                revert_i = input(f'Replay patch {self.__curr_release_s} [Y/n]? ') or 'Y'
+                if revert_i.upper() == 'Y':
+                    self.revert()
+                    force = True
+                else:
+                    sys.exit()
+            self.__patch(force=force)
+        self.__hop_cls.what_next()
+        return self.__curr_release_s
 
     def __register(self):
         "Mise à jour de la table half_orm_meta.hop_release"
         new_release = self.model.get_relation_class('half_orm_meta.hop_release')(
-            major=self.__release['major'],
-            minor=self.__release['minor'],
-            patch=int(self.__release['patch'])
+            major=self.__hop_cls.release['major'],
+            minor=self.__hop_cls.release['minor'],
+            patch=int(self.__hop_cls.release['patch'])
         )
         #FIXME
         commit = str(datetime.now())
@@ -163,24 +191,22 @@ class Patch:
         """
         if not os.path.isdir('./Backups'):
             os.mkdir('./Backups')
-        svg_file = self.__backup_path(self.__last_release_s)
+        svg_file = self.__backup_path(self.__hop_cls.last_release_s)
         if os.path.isfile(svg_file) and not force:
             sys.stderr.write(
-                f"Oops! there is already a dump for the {self.__last_release_s} release.\n")
+                f"Oops! there is already a dump for the {self.__hop_cls.last_release_s} release.\n")
             sys.stderr.write(f"Please use the --force option if you realy want to proceed.\n")
             sys.exit(1)
         subprocess.run(['pg_dump', self.dbname, '-f', svg_file], check=True)
 
-    def _patch(self, commit=None, force=False):
+    def __patch(self, commit=None, force=False):
         "Applies the patch and insert the information in the half_orm_meta.hop_release table"
         #TODO: simplify
-        last_release = self.get_current_release()
-        self.get_next_release(last_release)
-        if self.__release_s == '':
+        if self.__next_release is None:
             return
         # we've got a patch we switch to a new branch
         if not self.__hop_cls.model.production:
-            self.__hgit.set_branch(self.__release_s)
+            self.__hgit.set_branch(self.__hop_cls.release_s)
         self.save_database(force)
         if not os.path.exists(self.__patch_path):
             sys.stderr.write(f'The directory {self.__patch_path} does not exists!\n')
@@ -194,7 +220,7 @@ class Patch:
             sys.exit(1)
 
         if commit is None:
-            commit = self.__hgit.get_sha1_commit(changelog_file)
+            commit = self.__hgit.commit.hexsha
             if not force:
                 self.__hgit.exit_if_repo_is_not_clean()
 
@@ -245,17 +271,7 @@ class Patch:
 
     # def apply_issue(self, issue, commit=None, bundled_issue=None):
     #     "Applique un issue"
-    #     self._patch('devel/issues/{}'.format(issue), commit, bundled_issue)
-
-    def get_current_release(self):
-        """Returns the current release (dict)
-        """
-        return next(self.model.get_relation_class('half_orm_meta.view.hop_last_release')().select())
-
-    def get_release_s(cls, release):
-        """Returns the current release (str)
-        """
-        return '{major}.{minor}.{patch}'.format(**release)
+    #     self.__patch('devel/issues/{}'.format(issue), commit, bundled_issue)
 
     def prep_next_release(self, release_level):
         """Returns the next (major, minor, patch) tuple according to the release_level
@@ -263,7 +279,7 @@ class Patch:
         Args:
             release_level (str): one of ['patch', 'minor', 'major']
         """
-        current = self.get_current_release()
+        current = self.__hop_cls.get_current_db_release()
         next = {}
         next['major'] = current['major']
         next['minor'] = current['minor']
@@ -273,42 +289,19 @@ class Patch:
             next['minor'] = next['patch'] = 0
         if release_level == 'minor':
             next['patch'] = 0
+        new_release_s = '{major}.{minor}.{patch}'.format(**next)
+        print(f'PREPARING: {new_release_s}')
         patch_path = 'Patches/{major}/{minor}/{patch}'.format(**next)
         if not os.path.exists(patch_path):
-            changelog_msg = input('CHANGELOG message: ')
+            changelog_msg = input('CHANGELOG message (leave empty to abort): ')
+            if not changelog_msg:
+                print('Aborting')
+                return
             os.makedirs(patch_path)
             with open(f'{patch_path}/CHANGELOG.md', 'w', encoding='utf-8') as changelog:
                 changelog.write(changelog_msg)
-
-    def get_next_release(self, last_release=None):
-        "Renvoie en fonction de part le numéro de la prochaine release"
-        if last_release is None:
-            last_release = self.get_current_release()
-            msg = "CURRENT RELEASE: {major}.{minor}.{patch} at {time}"
-            if 'date' in last_release:
-                msg = "CURRENT RELEASE: {major}.{minor}.{patch}: {date} at {time}"
-            print(msg.format(**last_release))
-        self.__last_release_s = '{major}.{minor}.{patch}'.format(**last_release)
-        to_zero = []
-        tried = []
-        for part in ['patch', 'minor', 'major']:
-            next_release = dict(last_release)
-            next_release[part] = last_release[part] + 1
-            for sub_part in to_zero:
-                next_release[sub_part] = 0
-            to_zero.append(part)
-            next_release_path = '{major}/{minor}/{patch}'.format(**next_release)
-            next_release_s = self.get_release_s(next_release)
-            tried.append(next_release_s)
-            if os.path.exists('Patches/{}'.format(next_release_path)):
-                print(f"NEXT RELEASE: {next_release_s}")
-                self.__release = next_release
-                self.__release_s = next_release_s
-                self.__release_path = next_release_path
-                return next_release
-        print(f"No new release to apply after {self.__last_release_s}.")
-        print(f"Next possible releases: {', '.join(tried)}.")
-        return None
+        self.__hgit.set_branch(new_release_s)
+        print(f'You can now add your patch scripts (*.py, *.sql) in {patch_path}. See Patches/README.')
 
     def __add_relation(self, sql_dir, fqtn):
         with open(f'{sql_dir}/{fqtn}.sql', encoding='utf-8') as cmd:
@@ -328,7 +321,7 @@ class Patch:
         release_issue = self.model.has_relation('half_orm_meta.hop_release_issue')
         patch_confict = release or last_release or release_issue or penultimate_release
         if patch_confict:
-            release = self.get_release_s(self.get_current_release())
+            release = self.__hop_cls.get_release_s(self.__hop_cls.get_current_db_release())
             if release != '0.0.0':
                 sys.stderr.write('WARNING!\n')
                 sys.stderr.write(f'The hop patch system is already present at {release}!\n')
@@ -347,6 +340,4 @@ class Patch:
             "insert into half_orm_meta.hop_release values " +
             "(0,0,0, '', 0, now(), now(),'[0.0.0] First release', " +
             f'{date.today()})')
-
-        print("Patch system initialized at release '0.0.0'.")
         return "0.0.0"
